#include "VideoCapture.h"

CVideoCapture::CVideoCapture()
{
  pGraph = NULL;
  pBuilder = NULL;
  pControl = NULL;
  pCaptureSource = NULL;
  pGrabber = NULL;
  pSampleGrabber = NULL;
  pRenderer = NULL;
}

CVideoCapture::~CVideoCapture()
{
  SAFE_RELEASE(pGraph);
  SAFE_RELEASE(pBuilder);
  SAFE_RELEASE(pControl);
  SAFE_RELEASE(pCaptureSource);
  SAFE_RELEASE(pGrabber);
  SAFE_RELEASE(pSampleGrabber);
  SAFE_RELEASE(pRenderer);
}

BOOL CVideoCapture::InitGraph()
{
  // Create the Filter Graph Manager.
  hr = ::CoCreateInstance(CLSID_FilterGraph,
      NULL, CLSCTX_INPROC,
      IID_IGraphBuilder,
      (void **)&pGraph);
  if(hr != S_OK)
  {
    return FALSE;
  }
  // Create the VideoCapture Graph Builder.
  hr = ::CoCreateInstance(CLSID_CaptureGraphBuilder2,
      NULL,
      CLSCTX_INPROC,
      IID_ICaptureGraphBuilder2,
      (void **)&pBuilder);
  if(hr != S_OK)
  {
    return FALSE;
  }
  // Associate the graph with the builder.
  hr = pBuilder->SetFiltergraph(pGraph);
  if(hr != S_OK)
  {
    return FALSE;
  }
  // control
  hr = pGraph->QueryInterface(IID_IMediaControl,
          (void **)&pControl);
  return (hr == S_OK);
}

/**
* Add video source filter for specified device
*/
BOOL CVideoCapture::AddSource(LPTSTR bstrName)
{
  BOOL Result = FALSE;
  // Create the System Device Enumerator.
  ICreateDevEnum *pSysDevEnum;
  hr = ::CoCreateInstance(CLSID_SystemDeviceEnum,
    NULL,
    CLSCTX_INPROC_SERVER,
    IID_ICreateDevEnum,
    (void **)&pSysDevEnum);
  if(hr != S_OK)
  {
    return FALSE;
  }
  // Obtain a class enumerator for the video compressor category.
  IEnumMoniker *pEnumCat;
  hr = pSysDevEnum->CreateClassEnumerator(
    CLSID_VideoInputDeviceCategory,
    &pEnumCat,
    0);
  if(hr != S_OK)
  {
    return FALSE;
  }
  // Enumerate the monikers.
  ULONG cFetched = 0;
  IPropertyBag *pPropBag;
  IMoniker *pMoniker;
  VARIANT varName;

  while(!Result && pEnumCat->Next(1, &pMoniker, &cFetched) == S_OK)
  {
    if(pMoniker->BindToStorage(0, 0, IID_IPropertyBag, (void **)&pPropBag) == S_OK)
    {
      VariantInit(&varName);
      hr = pPropBag->Read(_T("FriendlyName"), &varName, 0);
      if(hr == S_OK)
      {
        if(_tcscmp(bstrName, varName.bstrVal) == 0)
        {
          hr = pMoniker->BindToObject(0, 0, IID_IBaseFilter, (void**)&pCaptureSource);
          hr = pGraph->AddFilter(pCaptureSource, _T("VideoCapture"));
          Result = TRUE;
          VariantClear(&varName);
        }
        hr = pPropBag->Release();
      }
    }
    hr = pMoniker->Release();
    pMoniker = NULL;
  }
  // clean up
  hr = pEnumCat->Release();
  hr = pSysDevEnum->Release();
  return Result;
}

BOOL CVideoCapture::AddRenderer()
{
  // Create NullRenderer
  hr = ::CoCreateInstance(CLSID_NullRenderer,
      NULL,
      CLSCTX_INPROC_SERVER,
      IID_IBaseFilter,
      (void **)&pRenderer);
  if(hr != S_OK)
  {
    return FALSE;
  }
  hr = pGraph->AddFilter(pRenderer, _T("NullRenderer"));
  return (hr == S_OK);
}

BOOL CVideoCapture::AddGrabber(CCallback *vid_call)
{
  // Create the Sample Grabber.
  hr = ::CoCreateInstance(CLSID_SampleGrabber,
    NULL,
    CLSCTX_INPROC_SERVER,
    IID_IBaseFilter,
    (void **)&pSampleGrabber);
  if(hr != S_OK)
  {
    return FALSE;
  }

  hr = pGraph->AddFilter(pSampleGrabber,
    _T("SampleGrabber"));
  if(hr != S_OK)
  {
    return FALSE;
  }

  // Settings
  hr = pSampleGrabber->QueryInterface(IID_ISampleGrabber,
    (void **)&pGrabber);
  if(hr != S_OK)
  {
    return FALSE;
  }

  // Set the media type.
  AM_MEDIA_TYPE mt;
  ZeroMemory(&mt, sizeof(AM_MEDIA_TYPE));
  mt.majortype = MEDIATYPE_Video;
  mt.subtype = MEDIASUBTYPE_RGB24;

  while(TRUE)
  {
    hr = pGrabber->SetMediaType(&mt);
    if(hr != S_OK)
    {
      break;
    }

    hr = pGrabber->SetOneShot(FALSE);
    if(hr != S_OK)
    {
      break;
    }
    hr = pGrabber->SetBufferSamples(TRUE);
    if(hr != S_OK)
    {
      break;
    }
    hr = pGrabber->SetCallback(vid_call, 1L);
    break;
  }

  SAFE_RELEASE(pGrabber);
  return (hr == S_OK);
}

BOOL CVideoCapture::Build(CCallback *vid_call, LPTSTR bstrName, int Width, int Height, int Rate)
{
  if(!InitGraph())
  {
    return FALSE;
  }
  if(bstrName == NULL)
  {
    return FALSE;
  }
  if(!AddSource(bstrName))
  {
    return FALSE;
  }
  if(!SetFormat(Width, Height, Rate))
  {
    return FALSE;
  }
  if(!AddGrabber(vid_call))
  {
    return FALSE;
  }
  if(!AddRenderer())
  {
    return FALSE;
  }

  hr = pBuilder->RenderStream(&PIN_CATEGORY_CAPTURE, &MEDIATYPE_Video,
        pCaptureSource, pSampleGrabber, pRenderer);

  return (hr == S_OK);
}

BOOL CVideoCapture::Start()
{
  if(!pControl)
  {
    return FALSE;
  }

  hr = pControl->Run();
  return SUCCEEDED(hr);
}

BOOL CVideoCapture::Stop()
{
  if(!pControl)
  {
    return FALSE;
  }

  hr = pControl->Stop();
  return (hr == S_OK);
}

BOOL CVideoCapture::SetFormat(int Width, int Height, int Rate)
{
  BOOL bResult = FALSE;
  AM_MEDIA_TYPE *pAMT = NULL;
  IAMStreamConfig *pConfig = NULL;

  hr = pBuilder->FindInterface(&PIN_CATEGORY_CAPTURE, &MEDIATYPE_Video,
          pCaptureSource, IID_IAMStreamConfig, (void **)&pConfig);
  if(hr != S_OK)
  {
    return FALSE;
  }

  hr = pConfig->GetFormat(&pAMT);
  if(hr == S_OK)
  {
    LONGLONG FrameRate = 10000000ULL / Rate;
    if(pAMT->formattype == FORMAT_VideoInfo)
    {
      VIDEOINFOHEADER *vih = (VIDEOINFOHEADER *)pAMT->pbFormat;
      vih->bmiHeader.biWidth = Width;
      vih->bmiHeader.biHeight = Height;
      vih->AvgTimePerFrame = FrameRate;
      hr = pConfig->SetFormat(pAMT);
      bResult = (hr == S_OK);
    }
    else
    {
      if(pAMT->formattype == FORMAT_VideoInfo2)
      {
        VIDEOINFOHEADER2 *vih = (VIDEOINFOHEADER2 *)pAMT->pbFormat;
        vih->bmiHeader.biWidth = Width;
        vih->bmiHeader.biHeight = Height;
        vih->AvgTimePerFrame = FrameRate;
        hr = pConfig->SetFormat(pAMT);
        bResult = (hr == S_OK);
      }
    }
    DeleteMediaType(pAMT);
  }

  SAFE_RELEASE(pConfig);
  return bResult;
}

void WINAPI CVideoCapture::DeleteMediaType(AM_MEDIA_TYPE *pmt)
{
  if(pmt)
  {
    FreeMediaType(*pmt);
    CoTaskMemFree((PVOID)pmt);
  }
}

void WINAPI CVideoCapture::FreeMediaType(AM_MEDIA_TYPE& mt)
{
  if (mt.cbFormat != 0)
  {
    CoTaskMemFree((PVOID)mt.pbFormat);
    // Strictly unnecessary but tidier
    mt.cbFormat = 0;
    mt.pbFormat = NULL;
  }
  SAFE_RELEASE(mt.pUnk);
}

BOOL CVideoCapture::VideoProperty(HWND hWnd)
{
  CAUUID caGUID;
  FILTER_INFO FilterInfo;
  ISpecifyPropertyPages *pSpecify;

  if(!pCaptureSource)
  {
    return FALSE;
  }

  // Discover if this filter contains a property page
  hr = pCaptureSource->QueryInterface(IID_ISpecifyPropertyPages,(void **)&pSpecify);
  if(hr != S_OK)
  {
    return FALSE;
  }

  hr = pCaptureSource->QueryFilterInfo(&FilterInfo);
  if (hr != S_OK)
  {
    return FALSE;
  }

  hr = pSpecify->GetPages(&caGUID);
  if (hr != S_OK)
  {
    return FALSE;
  }

  SAFE_RELEASE(pSpecify);

  // Display the filter's property page
  OleCreatePropertyFrame(
    hWnd,                   // Parent window
    0,                      // x (Reserved)
    0,                      // y (Reserved)
    FilterInfo.achName,     // Caption for the dialog box
    1,                      // Number of filters
    (IUnknown **)&pCaptureSource,   // Pointer to the filter
    caGUID.cElems,          // Number of property pages
    caGUID.pElems,          // Pointer to property page CLSIDs
    0,                      // Locale identifier
    0,                      // Reserved
    NULL                    // Reserved
  );

  CoTaskMemFree(caGUID.pElems);
  SAFE_RELEASE(FilterInfo.pGraph);

  return TRUE;
}

